"""Client for Open Exchange Rates API.

Examples:

```python
import openexchangerates as oxr

client = oxr.Client("YOUR_APP_ID")

# Get the latest exchange rates.
latest = client.latest("USD")

# Get historical exchange rates.
historical = client.historical("2021-01-01")

# Get time series exchange rates.
time_series = client.time_series(dt.date(2021, 1, 1), dt.date(2021, 1, 31))

# Get OHLC exchange rates.
ohlc = client.ohlc(dt.date(2021, 1, 1), dt.date(2021, 1, 31))

# Convert currency.
conversion = client.convert("USD", "EUR", 100)
```
"""

from __future__ import annotations

from oxr.client import Client

__all__ = ["Client"]



from __future__ import annotations

from typing import Literal

from typing_extensions import TypeAlias

# Valid API endpoints.
Endpoint: TypeAlias = Literal[
    "latest",
    "historical",
    "currencies",
    "convert",
    "time-series",
    "ohlc",
    "usage",
]


# Valid periods for the OHLC endpoint.
Period: TypeAlias = Literal["1m", "5m", "15m", "30m", "1h", "12h", "1d", "1w", "1mo"]

# Valid currency codes.
Currency: TypeAlias = str



from __future__ import annotations

from typing import Final

from oxr.exceptions import (
    Error,
    InvalidAppID,
    InvalidCurrency,
    InvalidDate,
    InvalidDateRange,
    NoAccessError,
)

_EXCEPTION_MAPPING: Final[dict[tuple[int, str], type[Error]]] = {
    (403, "not_allowed"): NoAccessError,
    (401, "invalid_app_id"): InvalidAppID,
    (400, "invalid_currency"): InvalidCurrency,
    (400, "invalid_date"): InvalidDate,
    (400, "invalid_date_range"): InvalidDateRange,
}


def get(code: int, message: str) -> type[Error] | None:
    """Get the error class for the given code and message."""
    return _EXCEPTION_MAPPING.get((code, message), None)



from __future__ import annotations

import datetime as dt
from decimal import Decimal

import pytest
import responses
from responses import matchers

import oxr

_BASE_RESPONSE = {
    "disclaimer": "https://openexchangerates.org/terms/",
    "license": "https://openexchangerates.org/license/",
    "timestamp": 1717622400,
}


@pytest.fixture
def client() -> oxr.Client:
    return oxr.Client("app_id")


@responses.activate
def test_latest(client: oxr.Client) -> None:
    resp = {
        "base": "USD",
        "rates": {"EUR": 0.85, "JPY": 110.0},
        **_BASE_RESPONSE,
    }
    responses.get(
        "https://openexchangerates.org/api/latest.json",
        json=resp,
        match=[
            matchers.query_param_matcher(
                {"app_id": "app_id", "base": "USD", "show_alternative": False}
            )
        ],
    )
    resp = client.latest()
    assert resp.base == "USD"
    assert resp.rates == {"EUR": Decimal("0.85"), "JPY": Decimal("110.0")}


@responses.activate
def test_historical(client: oxr.Client) -> None:
    resp = {
        "base": "USD",
        "rates": {"EUR": 0.85, "JPY": 110.0},
        **_BASE_RESPONSE,
    }
    responses.get(
        "https://openexchangerates.org/api/historical/2021-01-01.json",
        json=resp,
        match=[
            matchers.query_param_matcher(
                {
                    "app_id": "app_id",
                    "base": "USD",
                    "show_alternative": False,
                    "symbols": "EUR,JPY",
                }
            )
        ],
    )
    resp = client.historical(dt.date(2021, 1, 1), symbols=["EUR", "JPY"])
    assert resp.base == "USD"
    assert resp.rates == {"EUR": Decimal("0.85"), "JPY": Decimal("110.0")}


@responses.activate
def test_time_series(client: oxr.Client) -> None:
    resp = {
        "base": "USD",
        "start_date": "2021-01-01",
        "end_date": "2021-01-31",
        "rates": {
            "2021-01-01": {"EUR": 0.85, "JPY": 110.0},
            "2021-01-02": {"EUR": 0.86, "JPY": 111.0},
        },
        **_BASE_RESPONSE,
    }
    responses.get(
        "https://openexchangerates.org/api/time-series.json",
        json=resp,
        match=[
            matchers.query_param_matcher(
                {
                    "app_id": "app_id",
                    "base": "USD",
                    "show_alternative": False,
                    "symbols": "EUR,JPY",
                    "start": "2021-01-01",
                    "end": "2021-01-31",
                }
            )
        ],
    )
    resp = client.time_series(dt.date(2021, 1, 1), dt.date(2021, 1, 31), symbols=["EUR", "JPY"])
    assert resp.start_date == dt.date(2021, 1, 1)
    assert resp.end_date == dt.date(2021, 1, 31)
    assert resp.base == "USD"
    assert resp.rates == {
        dt.date(2021, 1, 1): {"EUR": Decimal("0.85"), "JPY": Decimal("110.0")},
        dt.date(2021, 1, 2): {"EUR": Decimal("0.86"), "JPY": Decimal("111.0")},
    }


@responses.activate
def test_convert(client: oxr.Client) -> None:
    resp = {
        "query": {"from": "USD", "to": "EUR", "amount": 100.0},
        "meta": {"rate": 0.85, "timestamp": 1717622400},
        "response": 85.0,
        **_BASE_RESPONSE,
    }
    responses.get(
        "https://openexchangerates.org/api/convert.json",
        json=resp,
        match=[
            matchers.query_param_matcher(
                {"app_id": "app_id", "from": "USD", "to": "EUR", "amount": 100.0}
            )
        ],
    )
    resp = client.convert(100, "USD", "EUR")
    assert resp.meta.rate == Decimal("0.85")
    assert resp.amount == Decimal("85.0")
    assert resp.meta.time == dt.datetime(2024, 6, 5, 21, 20, tzinfo=dt.timezone.utc)


@responses.activate
def test_ohlc(client: oxr.Client) -> None:
    resp = {
        "base": "USD",
        "start_time": "2021-01-01T00:00:00",
        "end_time": "2021-01-01T23:59:59Z",
        "rates": {"USD": {"open": 0.85, "high": 0.86, "low": 0.84, "close": 0.85, "average": 0.85}},
        **_BASE_RESPONSE,
    }
    responses.get(
        "https://openexchangerates.org/api/ohlc.json",
        json=resp,
        match=[
            matchers.query_param_matcher(
                {
                    "app_id": "app_id",
                    "base": "USD",
                    "show_alternative": False,
                    "period": "1d",
                    "start_time": "2021-01-01T00:00:00",
                    "symbols": "USD",
                }
            )
        ],
    )
    resp = client.ohlc(dt.datetime(2021, 1, 1), "1d", symbols=["USD"])
    assert resp.base == "USD"
    assert resp.start_time == dt.datetime(2021, 1, 1, 0, 0)
    assert resp.rates["USD"].open == Decimal("0.85")
    assert resp.rates["USD"].high == Decimal("0.86")
    assert resp.rates["USD"].low == Decimal("0.84")
    assert resp.rates["USD"].close == Decimal("0.85")
    assert resp.rates["USD"].average == Decimal("0.85")



"""Test endpoints that don't require a paid plan."""

from __future__ import annotations

import datetime as dt
import os

import pytest

import oxr
import oxr.exceptions


@pytest.fixture
def client() -> oxr.Client:
    app_id = os.getenv("OXR_APP_ID")
    if not app_id:
        pytest.skip("OXR_APP_ID is not set")
    return oxr.Client(app_id)


def test_latest(client: oxr.Client) -> None:
    resp = client.latest()
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert "JPY" in resp.rates

    # Narrow the symbols.
    resp = client.latest(symbols=["EUR"])
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert len(resp.rates) == 1


def test_historical(client: oxr.Client) -> None:
    resp = client.historical(dt.date(2021, 1, 1))
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert "JPY" in resp.rates

    # Narrow the symbols.
    resp = client.historical(dt.date(2021, 1, 1), symbols=["EUR"])
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert len(resp.rates) == 1


def test_currencies(client: oxr.Client) -> None:
    resp = client.currencies()
    assert "USD" in resp
    assert "EUR" in resp
    assert "JPY" in resp
    assert "ZAR" in resp


def test_usage(client: oxr.Client) -> None:
    resp = client.usage()
    assert "status" in resp
    assert "plan" in resp["data"]
    assert "usage" in resp["data"]
    assert "daily_average" in resp["data"]["usage"]


def test_no_access(client: oxr.Client) -> None:
    with pytest.raises(oxr.exceptions.NoAccessError):
        client.ohlc(dt.datetime(2021, 1, 1), "1d", symbols=["USD"])



from __future__ import annotations

import datetime as dt
from decimal import Decimal
from typing import Any, Iterable

import pydantic
import requests

from oxr import _exceptions, exceptions, responses
from oxr._base import BaseClient
from oxr._types import Currency, Endpoint, Period


class Client(BaseClient):
    """A client for the Open Exchange Rates API."""

    def _get(
        self,
        endpoint: Endpoint,
        query_params: dict[str, Any],
        path_params: list[str] | None = None,
    ) -> dict[str, Any]:
        """Make a GET request to the API."""
        url = self._prepare_url(endpoint, path_params)
        response = requests.get(
            url,
            params={"app_id": self._app_id, **query_params},
            timeout=self._timeout,
        )
        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            msg = response.json().get("message", "")
            exc = _exceptions.get(response.status_code, msg)
            if exc is not None:
                raise exc from error
            raise exceptions.Error(error) from None  # pragma: no cover

        return response.json()

    def currencies(self) -> responses.Currencies:
        currency_validator = pydantic.TypeAdapter(responses.Currencies)
        return currency_validator.validate_python(self._get("currencies", {}))

    def latest(
        self,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.Rates:
        params = {"base": base or self._base, "show_alternative": show_alternative}
        if symbols is not None:
            params["symbols"] = ",".join(symbols)

        self._get("latest", params)
        return responses.Rates.model_validate(self._get("latest", params))

    def historical(
        self,
        date: dt.date,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.Rates:
        params = {
            "base": base or self._base,
            "show_alternative": show_alternative,
        }
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        return responses.Rates.model_validate(
            self._get("historical", params, path_params=[date.isoformat()]),
        )

    def convert(
        self,
        amount: float | Decimal,
        from_: str,
        to: str,
    ) -> responses.Conversion:
        params = {"from": from_, "to": to, "amount": float(amount)}
        return responses.Conversion.model_validate(self._get("convert", params))

    def time_series(
        self,
        start: dt.date,
        end: dt.date,
        symbols: Iterable[Currency] | None = None,
        base: str | None = None,
        show_alternative: bool = False,
    ) -> responses.TimeSeries:
        params = {
            "start": start.isoformat(),
            "end": end.isoformat(),
            "show_alternative": show_alternative,
        }
        params["base"] = base or self._base
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        return responses.TimeSeries.model_validate(self._get("time-series", params))

    def ohlc(
        self,
        start_time: dt.datetime,
        period: Period,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.OHLC:
        params = {
            "start_time": start_time.isoformat(),
            "period": period,
            "show_alternative": show_alternative,
        }
        params["base"] = base or self._base
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        return responses.OHLC.model_validate(self._get("ohlc", params))

    def usage(self) -> dict[str, Any]:
        return self._get("usage", {})



from __future__ import annotations

import datetime as dt
from decimal import Decimal
from typing import Mapping, final

import pydantic
from typing_extensions import TypeAlias

from oxr._types import Currency


class _FrozenModel(pydantic.BaseModel):
    """A frozen Pydantic model."""

    model_config = {"frozen": True}


@final
class Rates(_FrozenModel):
    """The response for the latest and historical endpoints."""

    time: dt.datetime = pydantic.Field(validation_alias="timestamp")
    base: Currency
    rates: Mapping[Currency, Decimal] = pydantic.Field(
        examples=[{"EUR": Decimal("1.0"), "USD": Decimal("1.1")}]
    )


@final
class ConversionMetadata(_FrozenModel):
    """The metadata for the conversion response."""

    time: dt.datetime = pydantic.Field(validation_alias="timestamp")
    rate: Decimal


@final
class Conversion(_FrozenModel):
    """The response for the convert endpoint."""

    meta: ConversionMetadata
    amount: Decimal = pydantic.Field(validation_alias="response")


@final
class TimeSeries(_FrozenModel):
    """The response for the time series endpoint."""

    start_date: dt.date
    end_date: dt.date
    base: Currency
    rates: Mapping[dt.date, Mapping[Currency, Decimal]]


@final
class OHLCRates(_FrozenModel):
    """Response for rates as a part of the olch endpoint."""

    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    average: Decimal


@final
class OHLC(_FrozenModel):
    """The response for the olhc endpoint."""

    start_time: dt.datetime
    end_time: dt.datetime
    base: Currency
    rates: Mapping[Currency, OHLCRates]


Currencies: TypeAlias = Mapping[Currency, str]



from __future__ import annotations

import datetime as dt
import urllib.parse
from decimal import Decimal
from typing import Any

import pytest
from aioresponses import aioresponses

import oxr.asynchronous

_BASE_RESPONSE = {
    "disclaimer": "https://openexchangerates.org/terms/",
    "license": "https://openexchangerates.org/license/",
    "timestamp": 1717622400,
}


@pytest.fixture
def client() -> oxr.asynchronous.Client:
    return oxr.asynchronous.Client("app_id")


def _construct_url(base_url: str, params: dict[str, Any]) -> str:
    query_string = urllib.parse.urlencode(params)
    return f"{base_url}?{query_string}"


@pytest.mark.asyncio
async def test_latest(client: oxr.asynchronous.Client):
    resp = {
        "base": "USD",
        "rates": {"EUR": 0.85, "JPY": 110.0},
        **_BASE_RESPONSE,
    }
    url = _construct_url(
        "https://openexchangerates.org/api/latest.json",
        {"app_id": "app_id", "base": "USD", "show_alternative": "false"},
    )
    with aioresponses() as m:
        m.get(url, payload=resp)  # type: ignore
        response = await client.latest()
        assert response.base == "USD"
        assert response.rates == {"EUR": Decimal("0.85"), "JPY": Decimal("110.0")}


@pytest.mark.asyncio
async def test_historical(client: oxr.asynchronous.Client):
    resp = {
        "base": "USD",
        "rates": {"EUR": 0.85, "JPY": 110.0},
        **_BASE_RESPONSE,
    }
    url = _construct_url(
        "https://openexchangerates.org/api/historical/2021-01-01.json",
        {
            "app_id": "app_id",
            "base": "USD",
            "show_alternative": "false",
            "symbols": "EUR,JPY",
        },
    )
    with aioresponses() as m:
        m.get(url, payload=resp)  # type: ignore
        response = await client.historical(dt.date(2021, 1, 1), symbols=["EUR", "JPY"])
        assert response.base == "USD"
        assert response.rates == {"EUR": Decimal("0.85"), "JPY": Decimal("110.0")}


@pytest.mark.asyncio
async def test_time_series(client: oxr.asynchronous.Client):
    resp = {
        "base": "USD",
        "start_date": "2021-01-01",
        "end_date": "2021-01-31",
        "rates": {
            "2021-01-01": {"EUR": 0.85, "JPY": 110.0},
            "2021-01-02": {"EUR": 0.86, "JPY": 111.0},
        },
        **_BASE_RESPONSE,
    }
    url = _construct_url(
        "https://openexchangerates.org/api/time-series.json",
        {
            "app_id": "app_id",
            "base": "USD",
            "show_alternative": "false",
            "symbols": "EUR,JPY",
            "start": "2021-01-01",
            "end": "2021-01-31",
        },
    )
    with aioresponses() as m:
        m.get(url, payload=resp)  # type: ignore
        response = await client.time_series(
            dt.date(2021, 1, 1), dt.date(2021, 1, 31), symbols=["EUR", "JPY"]
        )
        assert response.start_date == dt.date(2021, 1, 1)
        assert response.end_date == dt.date(2021, 1, 31)
        assert response.base == "USD"
        assert response.rates[dt.date(2021, 1, 1)]["EUR"] == Decimal("0.85")
        assert response.rates[dt.date(2021, 1, 2)]["EUR"] == Decimal("0.86")
        assert response.rates[dt.date(2021, 1, 1)]["JPY"] == Decimal("110.0")
        assert response.rates[dt.date(2021, 1, 2)]["JPY"] == Decimal("111.0")


@pytest.mark.asyncio
async def test_convert(client: oxr.asynchronous.Client):
    resp = {
        "query": {"from": "USD", "to": "EUR", "amount": 100.0},
        "meta": {"rate": 0.85, "timestamp": 1717622400},
        "response": 85.0,
        **_BASE_RESPONSE,
    }
    url = _construct_url(
        "https://openexchangerates.org/api/convert.json",
        {"app_id": "app_id", "from": "USD", "to": "EUR", "amount": 100.0},
    )
    with aioresponses() as m:
        m.get(url, payload=resp)  # type: ignore
        response = await client.convert(100, "USD", "EUR")
        assert response.meta.rate == Decimal("0.85")
        assert response.amount == Decimal("85.0")
        assert response.meta.time == dt.datetime(2024, 6, 5, 21, 20, tzinfo=dt.timezone.utc)


@pytest.mark.asyncio
async def test_ohlc(client: oxr.asynchronous.Client):
    resp = {
        "base": "USD",
        "start_time": "2021-01-01T00:00:00",
        "end_time": "2021-01-01T23:59:59Z",
        "rates": {"USD": {"open": 0.85, "high": 0.86, "low": 0.84, "close": 0.85, "average": 0.85}},
        **_BASE_RESPONSE,
    }
    url = _construct_url(
        "https://openexchangerates.org/api/ohlc.json",
        {
            "app_id": "app_id",
            "base": "USD",
            "show_alternative": "false",
            "period": "1d",
            "start_time": "2021-01-01T00:00:00",
            "symbols": "USD",
        },
    )
    with aioresponses() as m:
        m.get(url, payload=resp)  # type: ignore
        response = await client.ohlc(dt.datetime(2021, 1, 1), "1d", symbols=["USD"])
        assert response.base == "USD"
        assert response.start_time == dt.datetime(2021, 1, 1, 0, 0)
        assert len(response.rates) == 1
        usd_rates = response.rates["USD"]
        assert usd_rates.open == Decimal("0.85")
        assert usd_rates.high == Decimal("0.86")
        assert usd_rates.low == Decimal("0.84")
        assert usd_rates.close == Decimal("0.85")
        assert usd_rates.average == Decimal("0.85")



from __future__ import annotations

import contextlib
import datetime as dt
from collections.abc import Iterable
from decimal import Decimal
from types import TracebackType
from typing import Any

import aiohttp
import pydantic
from typing_extensions import Self

from oxr import _exceptions, exceptions, responses
from oxr._base import BaseClient
from oxr._types import Currency, Endpoint, Period


def _encode_params(params: dict[str, Any]) -> dict[str, Any]:
    """yarl does not encode booleans as strings."""
    return {
        key: "true" if value is True else "false" if value is False else value
        for key, value in params.items()
    }


class Client(BaseClient, contextlib.AbstractAsyncContextManager["Client"]):
    """A asynchronous client for the Open Exchange Rates API."""

    _session: aiohttp.ClientSession | None = None

    async def _get(
        self,
        endpoint: Endpoint,
        query_params: dict[str, Any],
        path_params: list[str] | None = None,
    ) -> dict[str, Any]:
        url = self._prepare_url(endpoint, path_params)
        session = self._get_session()
        async with session.get(
            url,
            params=_encode_params({"app_id": self._app_id, **query_params}),
        ) as response:
            resp_json = await response.json()
            try:
                response.raise_for_status()
            except aiohttp.ClientResponseError as error:
                msg = resp_json.get("message", "")
                exc = _exceptions.get(error.status, msg)
                if exc is not None:
                    raise exc from error
                raise exceptions.Error(error) from None  # pragma: no cover
            return resp_json

    def _get_session(self) -> aiohttp.ClientSession:
        if self._session is None:
            timeout = aiohttp.ClientTimeout(total=self._timeout)
            self._session = aiohttp.ClientSession(timeout=timeout)
        return self._session

    async def currencies(self) -> responses.Currencies:
        """Get a list of available currencies."""
        currency_validator = pydantic.TypeAdapter(responses.Currencies)
        return currency_validator.validate_python(await self._get("currencies", {}))

    async def latest(
        self,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.Rates:
        """Get the latest exchange rates.

        Args:
            base: The base currency.
            symbols: The target currencies.
            show_alternative: Whether to show alternative currencies.
                Such as black market and digital currency rates.
        """
        params = {
            "app_id": self._app_id,
            "base": base or self._base,
            "show_alternative": show_alternative,
        }
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        rates = await self._get("latest", params)
        return responses.Rates.model_validate(rates)

    async def historical(
        self,
        date: dt.date,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.Rates:
        params = {
            "base": base or self._base,
            "show_alternative": show_alternative,
        }
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        return responses.Rates.model_validate(
            await self._get("historical", params, path_params=[date.isoformat()]),
        )

    async def convert(
        self,
        amount: float | Decimal,
        from_: str,
        to: str,
    ) -> responses.Conversion:
        params = {"from": from_, "to": to, "amount": float(amount)}
        return responses.Conversion.model_validate(
            await self._get("convert", params),
        )

    async def time_series(
        self,
        start: dt.date,
        end: dt.date,
        symbols: Iterable[Currency] | None = None,
        base: str | None = None,
        show_alternative: bool = False,
    ) -> responses.TimeSeries:
        params = {
            "start": start.isoformat(),
            "end": end.isoformat(),
            "show_alternative": show_alternative,
        }
        params["base"] = base or self._base
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        return responses.TimeSeries.model_validate(
            await self._get("time-series", params),
        )

    async def ohlc(
        self,
        start_time: dt.datetime,
        period: Period,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.OHLC:
        params = {
            "start_time": start_time.isoformat(),
            "period": period,
            "show_alternative": show_alternative,
        }
        params["base"] = base or self._base
        if symbols is not None:
            params["symbols"] = ",".join(symbols)
        return responses.OHLC.model_validate(
            await self._get("ohlc", params),
        )

    async def usage(self) -> dict[str, Any]:
        """Get the usage statistics for the API key."""
        return await self._get("usage", {})

    async def __aenter__(self) -> Self:
        """Entire client session."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        """Exit client session."""
        if self._session is not None:
            return await self.close()

    async def close(self) -> None:
        """Close the client session."""
        if self._session is not None:
            await self._session.close()
            self._session = None



from __future__ import annotations

from typing import ClassVar


class Error(Exception):
    """Base exception class for `openexchangerates`."""

    code: ClassVar[int | None] = None


class InvalidCurrency(Error):
    """Raised when an invalid currency is provided."""

    code = 400


class InvalidDate(Error):
    """Raised when an invalid date is provided."""

    code = 400


class InvalidDateRange(Error):
    """Raised when an invalid date range is provided."""

    code = 400


class InvalidAppID(Error):
    """Raised when an invalid App ID is provided."""

    code = 401


class NoAccessError(Error):
    """Raised when the App ID does not have access to the requested resource or endpoint."""

    code = 403



from __future__ import annotations

import abc
import datetime as dt
from collections.abc import Awaitable
from decimal import Decimal
from typing import Any, Final, Iterable

from oxr import responses
from oxr._types import Currency, Endpoint, Period

_BASE_URL: Final = "https://openexchangerates.org/api"


class BaseClient(abc.ABC):
    def __init__(
        self,
        app_id: str,
        *,
        base: Currency = "USD",
        base_url: str = _BASE_URL,
        timeout: float = 5,
    ) -> None:
        """Initialize the client.

        Args:
            app_id: The app id for the API.
            base: The base currency to use across the client.
            base_url: The base URL for the API.
                Unless you are using a custom endpoint, or for testing
                purposes, you should not need to change this.
            timeout: The request timeout in seconds.
        """
        self._app_id = app_id
        self._base = base
        self._base_url = base_url
        self._timeout = timeout

    def _prepare_url(self, endpoint: Endpoint, path_params: list[str] | None) -> str:
        base_url = f"{self._base_url}/{endpoint}"
        url_paths = "/".join(path_params) if path_params else ""
        return f"{base_url}/{url_paths}.json" if url_paths else f"{base_url}.json"

    @abc.abstractmethod
    def _get(
        self,
        endpoint: Endpoint,
        query_params: dict[str, Any],
        path_params: list[str] | None = None,
    ) -> dict[str, Any] | Awaitable[dict[str, Any]]:
        """Make a GET request to the API."""

    @abc.abstractmethod
    def latest(
        self,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.Rates | Awaitable[responses.Rates]:
        """Get the latest exchange rates.

        Args:
            base: The base currency.
            symbols: The target currencies.
            show_alternative: Whether to show alternative currencies.
                Such as black market and digital currency rates.
        """

    @abc.abstractmethod
    def currencies(self) -> responses.Currencies | Awaitable[responses.Currencies]:
        """Get a mapping of available currency codes to names."""

    @abc.abstractmethod
    def historical(
        self,
        date: dt.date,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.Rates | Awaitable[responses.Rates]:
        """Get historical exchange rates.

        Args:
            date: The date of the rates.
            base: The base currency.
            symbols: The target currencies.
            show_alternative: Whether to show alternative currencies.
                Such as black market and digital currency rates.
        """

    @abc.abstractmethod
    def convert(
        self,
        amount: float | Decimal,
        from_: str,
        to: str,
    ) -> responses.Conversion | Awaitable[responses.Conversion]:
        """Convert an amount between two currencies.

        Args:
            amount: The amount to convert.
            from_: The source currency.
            to: The target currency.
            date: The date of the rates to use.
        """

    @abc.abstractmethod
    def time_series(
        self,
        start: dt.date,
        end: dt.date,
        symbols: Iterable[Currency] | None = None,
        base: str | None = None,
        show_alternative: bool = False,
    ) -> responses.TimeSeries | Awaitable[responses.TimeSeries]:
        """Get historical exchange rates for a range of dates.

        Args:
            start: The start date of the range.
            end: The end date of the range.
            symbols: The target currencies.
            base: The base currency.
            show_alternative: Whether to show alternative currencies.
                Such as black market and digital currency rates.
        """

    @abc.abstractmethod
    def ohlc(
        self,
        start_time: dt.datetime,
        period: Period,
        base: str | None = None,
        symbols: Iterable[Currency] | None = None,
        show_alternative: bool = False,
    ) -> responses.OHLC | Awaitable[responses.OHLC]:
        """Get the latest open, low, high, and close rates for a currency.

        Args:
            base: The base currency.
            symbols: The target currencies.
            show_alternative: Whether to show alternative currencies.
                Such as black market and digital currency rates.
        """



"""Test endpoints that don't require a paid plan."""

from __future__ import annotations

import datetime as dt
import os

import pytest

import oxr.asynchronous
import oxr.exceptions


@pytest.fixture
def client() -> oxr.asynchronous.Client:
    app_id = os.getenv("OXR_APP_ID")
    if not app_id:
        pytest.skip("OXR_APP_ID is not set")
    return oxr.asynchronous.Client(app_id)


@pytest.mark.asyncio
async def test_latest(client: oxr.asynchronous.Client) -> None:
    resp = await client.latest()
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert "JPY" in resp.rates

    # Narrow the symbols.
    resp = await client.latest(symbols=["EUR"])
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert len(resp.rates) == 1


@pytest.mark.asyncio
async def test_historical(client: oxr.asynchronous.Client) -> None:
    resp = await client.historical(dt.date(2021, 1, 1))
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert "JPY" in resp.rates

    # Narrow the symbols.
    resp = await client.historical(dt.date(2021, 1, 1), symbols=["EUR"])
    assert resp.base == "USD"
    assert "EUR" in resp.rates
    assert len(resp.rates) == 1


@pytest.mark.asyncio
async def test_currencies(client: oxr.asynchronous.Client) -> None:
    resp = await client.currencies()
    assert "USD" in resp
    assert "EUR" in resp
    assert "JPY" in resp
    assert "ZAR" in resp


@pytest.mark.asyncio
async def test_usage(client: oxr.asynchronous.Client) -> None:
    resp = await client.usage()
    assert "status" in resp
    assert "plan" in resp["data"]
    assert "usage" in resp["data"]
    assert "daily_average" in resp["data"]["usage"]


@pytest.mark.asyncio
async def test_session_static(client: oxr.asynchronous.Client) -> None:
    async with client as async_client:
        session = async_client._session  # type: ignore
        assert session is None
        await async_client.usage()
        session = async_client._session  # type: ignore
        await async_client.currencies()
        assert async_client._session is session  # type: ignore


@pytest.mark.asyncio
async def test_no_access(client: oxr.asynchronous.Client):
    """Test that we properly re-raise 403 errors as `NoAccessError`."""
    with pytest.raises(oxr.exceptions.NoAccessError):
        await client.ohlc(dt.datetime(2021, 1, 1), "1d", symbols=["USD"])
